'''Sports FM - 跑步FM APIs'''
from . import EapiCryptoRequest
import json

@EapiCryptoRequest
def GetSportsFMRecommendations(limit=3,bpm:int = 50,e_r=True):
    '''移动端 - 小程序 - 获取跑步FM歌曲推荐

    Args:
        limit (int, optional): 获取数目. Defaults to 3.
        bpm (int): 实为‘每分钟步数’. Defaults to 50.
        e_r (bool, optional): 未知. Defaults to True.

    Returns:
        dict
    '''
    return '/eapi/radio/sport/get',{"limit":str(limit),"bpm":str(bpm),"e_r":str(e_r).lower()}

@EapiCryptoRequest
def GetCalculatedSportsFMStatus(distance=0,maxbpm=0,time=0,songList=[],steps=0,bpm=0,e_r=True):
    '''移动端 - 小程序 - 计算运动详情

    Args:
        distance (int, optional): 距离 - 单位未知. Defaults to 0.
        maxbpm (int, optional): 最大心率 - 暂存. Defaults to 0.
        time (int, optional): 运动用时 - 秒为单位. Defaults to 0.
        songList (list, optional): 播放过的歌曲列表. Defaults to [].
        steps (int, optional): 行进的步数. Defaults to 0.
        bpm (int, optional): 结束时刻心率. Defaults to 0.
        e_r (bool, optional): 未知. Defaults to True.

    Returns:
        dict
    '''
    return '/eapi/radio/sport/calculate',{
        "distance":str(distance),
        "maxbpm":str(maxbpm),
        "time":str(time),
        "songList":json.dumps(songList),
        "steps":str(steps),
        "bpm":str(bpm),
        "e_r":str(e_r).lower()
    }


'''用户 - User 有关 APIs'''
from . import WeapiCryptoRequest,UserIDBasedApi,LoginRequiredApi

@WeapiCryptoRequest
@UserIDBasedApi
def GetUserDetail(user_id = 0):
    '''网页端 - 获取某用户资料详情

    Args:
        user_id (int): 用户 ID。置 0 表示当前已登录的用户 . defaults to 0

    Returns:
        dict
    '''
    return '/weapi/v1/user/detail/%s' % user_id,{}

@WeapiCryptoRequest
@UserIDBasedApi
def GetUserPlaylists(user_id, offset=0, limit=1001):
    '''网页端 - 获取某用户创建的歌单

    Args:
        user_id (int): 用户 ID。置 0 表示当前已登录的用户 . defaults to 0
        offset (int, optional): 获取偏移数. Defaults to 0.
        limit (int, optional): 单次获取量. Defaults to 30.

    Returns:
        dict
    '''    
    return '/weapi/user/playlist',{'offset': str(offset), 'limit': str(limit),'uid': str(user_id)}

@WeapiCryptoRequest
@LoginRequiredApi
def GetUserArtistSubs(limit=30):
    '''网页端 - 获取收藏歌手内容
    Args:
        limit (int, optional): 单次获取量. Defaults to 30.

    Returns:
        dict
    '''
    return '/api/artist/sublist', {"limit": str(limit)}

SIGNIN_TYPE_MOBILE = 0
'''移动端签到 +4 EXP'''
SIGNIN_TYPE_WEB    = 1
'''网页端签到 +1 EXP'''

@WeapiCryptoRequest
@LoginRequiredApi
def SetSignin(type=0):
    '''移动端、PC端 - 每日签到

    Args:
        type (int, optional): 签到类型，请使用本模块内 SIGNIN_TYPE_... 之一 .Defaults to SIGNIN_TYPE_MOBILE.

    Returns:
        dict
    '''
    return '/weapi/point/dailyTask',{'type':str(type)}



'''# Watchman.js implementation'''
import json,math,ctypes
from hashlib import md5
from time import time

# region Constant strings
Sc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
Tc = '='
'''Base64 chars'''
Rb = 'LPc4uQNptC2A6y.R90DOBIroS+qnx/eb3FM8fW1UZG7VmwvksgjhaTKzlXdiYEHJ'
db = 5
'''Default obfuscation chars'''
WM_DID = 'Eyj1ZVEWep5ERQEFFVMuN0xNxLvKykXT' # fallback value
'''DeviceID - XXX - device Id - Probably won't be bothering with this anymore...'''
# endregion
# region Bitwise operations
'''These exsist beacues JS's bit handling is not correct arithmeticly - AND,OR,XOR,NOT,and LS,RS are few examples'''
def _int(v):return ctypes.c_int(v).value
def _lshift(v,bs):
    return _int(v << (bs & 0x1F))
def _rshift(v,bs):
    return _int(v >> (bs & 0x1F))
# Bit shifting only takes care of LS 5 bits
def _logical_rshift(v,bs):
    return _int( ( v % (0xFFFFFFFF + 1)) >> (bs & 0x1F))    
def _truncate(v):
    return v & 0xFFFFFFFF    
# MDN : All javascript bit operations are done on 32 bit integers
def _and(v1,v2): return _int( _truncate(v1) & _truncate(v2))
def  _or(v1,v2): return _int( _truncate(v1) | _truncate(v2))
def _xor(v1,v2): return _int( _truncate(v1) ^ _truncate(v2))
def _not(v1):    return _int(~_truncate(v1))
# endregion
# region MD5
def ha(b, c):
    '''Bitwise sum'''
    d = (b & 65535) + (c & 65535)
    br16 = _rshift(b,16)
    cr16 = _rshift(c,16)
    dr16 = _rshift(d,16)
    r = br16 + cr16 + dr16
    rl16 = _lshift(r,16)    
    return rl16 | d & 65535
def Yb(e):
    '''MD5 hasher'''
    from . import Crypto
    return Crypto.HashDigest(e)
def Zb(e):
    '''MD5 digest hexi-fier'''
    from . import Crypto
    return Crypto.HexDigest(e)
# endregion
# region Special functions
def C(e):
    '''Integer mapping - (-inf,inf) -> [-127,127]'''
    if e < -128:
        return C(128 - (-128 - e))
    if (e >= -128 and e <= 127):
        return e
    if (e > 127):
        return C(-129 + e - 127)
def X(b):
    '''Integer to bytes - int(32 bits) -> [4 bytes]'''
    return [C(b >> 24 & 255),C(b >> 16 & 255),C(b >> 8 & 255),C(b & 255)]
def Z(e):
    '''encodeURIComponent()'''
    return e
def ac(e):
    '''unescape()'''
    return e
def la(e):
    '''parseInt()'''
    return int(e)
def ya(e):    
    '''String to charcode array'''
    return [ord(char) for char in Z(e)] if isinstance(e,str) else e
def Ub(e):
    '''Char to hexdecimal [00,FF] string'''
    c = "0123456789abcdef"
    return c[e >> 4 & 15] + c[e & 15]
def gb(e):
    '''String to hexstring converter'''
    return ''.join([Ub(a) for a in e])
def La(e):    
    '''Simple XOR cipherer'''
    d = [31, 125, -12, 60, 32, 48]
    e = ya(e)
    g = []
    for q in range(0,len(e)):         
        g.append(C(_xor(e[q],d[q % len(d)])))
        g[q] = C(-g[q])
    return gb(g)
def ga(e, c, d, r, g):
    '''Array shuffler'''
    for h in range(0,g):d[r + h] = e[c + h]
def hb(e):
    '''Hexstring to integer array'''
    c = []
    for r in range(0,len(e),2):
        h = _lshift(int(e[r], 16),4)
        m = int(e[r + 1], 16)
        c.append(C(h + m))
    return c    
def Qb(e, c, d=0, r=Rb, g=db):
    '''Integer array to obfuscated string'''
    h, m,g = 0,[],str(g)
    def push(*a):m.extend(a)
    if d == 1:
        d = e[c];
        h = 0;
        push(r[ d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], g, g)
    elif d == 2:
        d = e[c];
        h = e[c + 1];
        e = 0;
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], g)
    elif d == 3:
        d = e[c];
        h = e[c + 1];
        e = e[c + 2];
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], r[e & 63])
    else:
        return None
    return ''.join(m)
def Sb(e, c=[], d=db):
    '''Characters mapped to another array'''
    r = 3
    h = 0
    g = []
    while (h < len(e)):
        if (h + r <= len(e)):
            g.append(Qb(e, h, r, c, d))
            h += r
        else:
            g.append(Qb(e, h, len(e) - h, c, d));
            break
    return "".join(g)
def Tb(a):
    return Sb(a,Sc,Tc)
def b():
    '''OTP tokenizer'''
    e = int(time() * 1000) # it seems that this timestamp should within 1mo of today's time
    c = math.floor(e / 4294967296)
    d = e % 4294967296
    e = X(c)
    d = X(d)
    c = [ 0 ] * 8    
    ga(e, 0, c, 0, 4);
    ga(d, 0, c, 4, 4);  
    d = [ 0 ] * 8 # should be random nums of [0-255] * 8
    e = [ 0 ] * 16
    for g in range(0,len(c) * 2):
        if (g % 2 == 0):
            f = int(g / 2)
            e[g] = e[g] | (d[f] & 16) >> 4 | (d[f] & 32) >> 3 | (d[f] & 64) >> 2 | (d[f] & 128) >> 1 | (c[f] & 16) >> 3 | (c[f] & 32) >> 2 | (c[f] & 64) >> 1 | (c[f] & 128) >> 0
        else:
            f = math.floor(g / 2)
            e[g] = e[g] | (d[f] & 1) << 0 | (d[f] & 2) << 1 | (d[f] & 4) << 2 | (d[f] & 8) << 3 | (c[f] & 1) << 1 | (c[f] & 2) << 2 | (c[f] & 4) << 3 | (c[f] & 8) << 4
        e[g] = C(e[g])
    c = gb(e)
    c = Zb(Yb(ac(Z(c + "dAWsBhCqtOaNLLJ25hBzWbqWXwiK99Wd"))))
    c = hb(c[:16])
    c = Tb(c+e)
    return c
def bc(c,f=1,g=WM_DID):
    '''Generates hash of our local config'''
    t = ac(Z(str(f) + str(g) + str(c) + "WoeTpXnDDPhiAvsJUUIY3RdAo2PKaVwi"))
    t = Yb(t)
    return La(json.dumps({
        'r': f,      # config.va - should be 1 all the time
        'd': g,      # g || "",device ID
        'b': c,
        't': Tb(t)
    }))
# end region    
def checkToken(deviceId=WM_DID):
    '''Generates `checkToken` parameter
    
    Args:
        deviceId (str, optional): XXX: Device ID fetched. Defaults to WM_DID.

    Returns:
        str
    '''
    return bc(b(),g=deviceId)


'''伪装，Anti-AntiCSRF 有关 APIs'''
from . import WeapiCryptoRequest
import json
@WeapiCryptoRequest
def MockWeblog(logs):
    '''网页端 - 用户足迹接口 - reserved

    Args:
        logs (Any): 记录的用户足迹

    Returns:
        dict
    '''
    return '/weapi/feedback/weblog',{'logs':json.dumps(logs)}


'''登录、Anti - AntiCSRF 有关 APIs'''
from . import EapiCryptoRequest, WeapiCryptoRequest,GetCurrentSession,logger,Crypto,LoginFailedException
import time

def WriteLoginInfo(response):
    '''写登录态入当前 Session

    Args:
        response (dict): 解码后的登录态

    Raises:
        LoginFailedException: 登陆失败时发生
    '''
    sess = GetCurrentSession()
    sess.login_info = {'tick': time.time(), 'content':response}
    if not sess.login_info['content']['code'] == 200:
        sess.login_info['success'] = False
        raise LoginFailedException(sess.login_info['content'])
    sess.login_info['success'] = True
    cookie = sess.cookies.get_dict()
    sess.csrf_token = cookie['__csrf']
    logger.debug('Updated login info for user %s' % sess.login_info['content']['profile']['nickname'])

@WeapiCryptoRequest
def LoginLogout():
    '''网页端 - 登出账号

    Returns:
        dict
    '''
    return '/weapi/logout',{}

@WeapiCryptoRequest
def LoginRefreshToken():
    '''网页端 - 刷新登录令牌

    Returns:
        dict
    '''
    return '/weapi/w/login/cellphone',{}

@WeapiCryptoRequest
def LoginQrcodeUnikey(type=1):
    '''网页端 - 获取二维码登录令牌

    - 该令牌UUID适用于以下 URL：
     - music.163.com/login?codekey={UUID}
    - 若需要使用该URL，应由网易云音乐移动客户端扫描其二维码 - 链接不会触发入口
    - 登录态将在 `LoginQrcodeCheck` 中更新，需周期性检测
    
    Args:
        type (int, optional): 未知. Defaults to 1.

    Returns:
        dict
    '''
    return '/weapi/login/qrcode/unikey',{"type":str(type)}

@WeapiCryptoRequest
def LoginQrcodeCheck(unikey,type=1):
    '''网页端 - 二维码登录状态检测

    Args:
        key (str): 二维码 unikey
        type (int, optional): 未知. Defaults to 1.

    Returns:
        dict
    '''
    return '/weapi/login/qrcode/client/login',{'key':str(unikey),'type':type}

@WeapiCryptoRequest
def LoginTypeSwitch():
    '''网页端 - 切换登录方式 - reserved

    Returns:
        dict
    '''
    return '/weapi/w/user/login/type/switch',{}

@WeapiCryptoRequest
def GetCurrentLoginStatus():
    '''网页端 - 获取当前登录态
    
    Returns:
        dict
    '''
    return '/weapi/w/nuser/account/get',{}

def LoginViaCellphone(phone='', password='',ctcode=86,remeberLogin=True) -> dict:
    '''网页端 - 手机号登陆

    Args:
        phone (str, optional): 手机号. Defaults to ''.
        password (str, optional): 明文密码. Defaults to ''.
        countrycode (int, optional): 国家代码. Defaults to 86.
        remeberLogin (bool, optional): 是否‘自动登录’，开启可延长 CSRF 令牌时效. Defaults to True.

    Raises:
        LoginFailedException: 登陆失败时发生

    Returns:
        dict
    '''
    path='/weapi/w/login/cellphone'
    sess = GetCurrentSession()
    if (phone and password):
        sess.phone = phone
        sess.password = password
        return LoginViaCellphone()
    else:
        if (hasattr(sess,'phone') and hasattr(sess,'password')):
            md5_password = Crypto.HashHexDigest(sess.password)
            login_status = WeapiCryptoRequest(lambda:(path,{
                "phone":str(sess.phone),
                "checkToken":Crypto.checkToken(),
                "password":str(md5_password),
                "rememberLogin":str(remeberLogin).lower(),
                "countrycode":str(ctcode),
            }))() 
            WriteLoginInfo(login_status)
            return sess.login_info
        else:
            raise LoginFailedException('Phone number or password not set')

@WeapiCryptoRequest
def SetSendRegisterVerifcationCodeViaCellphone(cell : str,ctcode=86):
    '''网页端 - 发送验证码

    - 验证码 24h 内最多发送五次

    Args:
        cell (str): 手机号
        ctcode (int, optional): 国家代码. Defaults to 86.

    Returns:
        dict
    '''
    return '/weapi/sms/captcha/sent',{'cellphone': str(cell),'ctcode': ctcode}
    

@WeapiCryptoRequest
def GetRegisterVerifcationStatusViaCellphone(cell : str,captcha:str,ctcode=86):
    '''网页端 - 检查验证码是否正确

    Args:
        cell (str): 手机号
        captcha (str): 验证码
        ctcode (int, optional): 国家代码. Defaults to 86.

    Returns:
        dict
    '''
    return '/weapi/sms/captcha/verify',{'cellphone': str(cell),'captcha':str(captcha),'ctcode': ctcode}

@WeapiCryptoRequest
def SetRegisterAccountViaCellphone(cell : str,captcha:str,nickname:str,password:str):
    '''网页端 - 手机号注册

    - 需要已通过 `SetSendRegisterVerifcationCodeViaCellphone` 发送验证码
    - `忘记密码` 同样使用该 API
    - 成功后，现`Session`将登陆进入该账号

    Args:
        cell (str): 手机号
        captcha (str): 验证码
        nickname (str): 昵称
        password (str): 密码

    Returns:
        dict
    '''
    return '/weapi/w/register/cellphone',{
        'captcha':str(captcha),
        'nickname':str(nickname),
        'password':Crypto.HashHexDigest(password),
        'phone':str(cell),
        'checkToken':Crypto.checkToken(),
    }    

@EapiCryptoRequest
def CheckIsCellphoneRegistered(cell : str,prefix=86):
    '''移动端 - 检查某手机号是否已注册

    Args:
        cell (str): 手机号
        prefix (int): 区号 . Defaults to 86

    Returns:
        dict
    '''
    return '/eapi/cellphone/existence/check',{'cellphone':cell,'countrycode':prefix}